 using System;
using System.Collections.Generic;
using System.Linq;
using Newtonsoft.Json;
using Oxide.Core;

 
namespace Oxide.Plugins
{
    [Info("Kits", "Reneb + ✬☯Kub4☯✬#1950", "1.0.4", ResourceId = 1494)]
    [Description("Create kits of items for players to use.")]

    class Kits : HurtworldPlugin
    {
        //////////////////////////////////////////////////////////////////////////////////////////
        ///// Plugin initialization
        //////////////////////////////////////////////////////////////////////////////////////////

        void Loaded()
        {
            LoadData();
            try
            {
                kitsData = Interface.Oxide.DataFileSystem.ReadObject<Dictionary<ulong, Dictionary<string, KitData>>>("Kits_Data");
            }
            catch
            {
                kitsData = new Dictionary<ulong, Dictionary<string, KitData>>();
            }
        }

        void OnServerInitialized() => InitializePermissions();

        const string permControl = "kits.use";

        void InitializePermissions()
        {
            foreach (var kit in storedData.Kits.Values)
            {
                if (string.IsNullOrEmpty(kit.permission)) continue;
                permission.RegisterPermission(kit.permission, this);
            }

            permission.RegisterPermission(permControl, this);
        }

        //////////////////////////////////////////////////////////////////////////////////////////
        ///// Configuration
        //////////////////////////////////////////////////////////////////////////////////////////

        void OnPlayerRespawn(PlayerSession session)
        {
            if (!storedData.Kits.ContainsKey("autokit")) return;

            var thereturn = Interface.Oxide.CallHook("canRedeemKit", session);

            if (thereturn == null)
            {
                var playerinv = session.WorldPlayerEntity.GetComponent<PlayerInventory>();

                for (var j = 0; j < playerinv.Capacity; j++)
                {
                    if (playerinv.Items[j] == null) continue;
                    if (playerinv.Items[j].Item == null) continue;
                    Singleton<ClassInstancePool>.Instance.ReleaseInstanceExplicit(playerinv.Items[j]);
                    playerinv.Items[j] = null;
                }
                GiveKit(session, "autokit");
            }
        }

        //////////////////////////////////////////////////////////////////////////////////////////
        ///// Kit Creator
        //////////////////////////////////////////////////////////////////////////////////////////

        List<KitItem> GetPlayerItems(PlayerSession session)
        {
            var playerinv = session.WorldPlayerEntity.GetComponent<PlayerInventory>();

            var kititems = new List<KitItem>();

            for (var i = 0; i < playerinv.Capacity; i++)
            {
                var item = playerinv.Items[i];

                if (item?.Item == null) continue;
                kititems.Add(new KitItem
                {
                    itemid = item.Item.ItemId,
                    amount = item.StackSize,
                    slot = i
                });
            }
            return kititems;
        }

        //////////////////////////////////////////////////////////////////////////////////////////
        ///// Kit Redeemer
        //////////////////////////////////////////////////////////////////////////////////////////

        private void TryGiveKit(PlayerSession session, string kitname)
        {
            var success = CanRedeemKit(session, kitname) as string;

            if (success != null)
            {
                hurt.SendChatMessage(session, null, success);
                return;
            }
            success = GiveKit(session, kitname) as string;
            if (success != null)
            {
                hurt.SendChatMessage(session, null, success);
                return;
            }
            hurt.SendChatMessage(session, null, "<color=#0066cc>Otrzymałeś zestaw!</color>");

            ProccessKitGiven(session, kitname);
        }

        void ProccessKitGiven(PlayerSession session, string kitname)
        {
            Kit kit;

            if (string.IsNullOrEmpty(kitname) || !storedData.Kits.TryGetValue(kitname, out kit)) return;

            var kitData = GetKitData(session.SteamId.m_SteamID, kitname);

            if (kit.max > 0) kitData.max += 1;

            if (kit.cooldown > 0) kitData.cooldown = CurrentTime() + kit.cooldown;
        }

        object GiveKit(PlayerSession session, string kitname)
        {
            Kit kit;

            if (string.IsNullOrEmpty(kitname) || !storedData.Kits.TryGetValue(kitname, out kit)) return $"Zestaw '{kitname}' nie istnieje";

            var playerinv = session.WorldPlayerEntity.GetComponent<PlayerInventory>();
            var amanager = Singleton<AlertManager>.Instance;
            var itemmanager = Singleton<GlobalItemManager>.Instance;

            foreach (var kitem in kit.items)
            {
                if (playerinv.Items[kitem.slot] == null)
                {
                    var item = itemmanager.GetItem(kitem.itemid);
                    var iitem = new ItemInstance(item, kitem.amount);
                    playerinv.Items[kitem.slot] = iitem;
                    amanager.ItemReceivedServer(iitem.Item, iitem.StackSize, session.Player);
                    playerinv.Invalidate(false);
                }
                else
                    itemmanager.GiveItem(session.Player, itemmanager.GetItem(kitem.itemid), kitem.amount);
            }
            return true;
        }

        //////////////////////////////////////////////////////////////////////////////////////////
        ///// Check Kits
        //////////////////////////////////////////////////////////////////////////////////////////

        bool isKit(string kitname) => !string.IsNullOrEmpty(kitname) && storedData.Kits.ContainsKey(kitname);

        static readonly DateTime epoch = new DateTime(1970, 1, 1, 0, 0, 0);
        static double CurrentTime() { return DateTime.UtcNow.Subtract(epoch).TotalSeconds; }

        bool CanSeeKit(PlayerSession session, string kitname, out string reason)
        {
            reason = string.Empty;

            Kit kit;

            if (string.IsNullOrEmpty(kitname) || !storedData.Kits.TryGetValue(kitname, out kit)) return false;
            if (kit.hide) return false;
            if (kit.authlevel > 0)
                if (!session.IsAdmin) return false;
            if (!string.IsNullOrEmpty(kit.permission))
                if (!permission.UserHasPermission(session.SteamId.ToString(), kit.permission)) return false;
            if (kit.max > 0)
            {
                var left = GetKitData(session.SteamId.m_SteamID, kitname).max;

                if (left >= kit.max)
                {
                    reason += "- 0 pozostało";
                    return false;
                }
                reason += $"- {(kit.max - left)} pozostało";
            }
            if (kit.cooldown > 0)
            {
                var cd = Convert.ToInt32(GetKitData(session.SteamId.m_SteamID, kitname).cooldown);
                var ct = Convert.ToInt32(CurrentTime());

                if (cd > ct && cd != 0.0)
                {
                    int a;
                    int t;
                    int s = cd - ct;

                    a = s / 3600;
                    t = s % 3600;
                    if (t != 0)
                        t /= 60;
                    if (s < 3600)
                    {
                        reason += $"| {t} minut/y";
                        return false;
                    }
                    if (s >= 3600 && t != 0)
                    {
                        reason += $"| {a} godzin/y {t} minut/y";
                    }
                    else
                    {
                        reason += $"| {a} godzin/y";
                    }

                    return false;
                }
            }
            return true;
        }

        object CanRedeemKit(PlayerSession session, string kitname)
        {
            Kit kit;

            if (string.IsNullOrEmpty(kitname) || !storedData.Kits.TryGetValue(kitname, out kit)) return $"Kit '{kitname}' nie istnieje";

            var thereturn = Interface.Oxide.CallHook("canRedeemKit", session);

            if (thereturn != null)
            {
                if (thereturn is string) return thereturn;
                return "Nie możesz teraz odebrać zestawu.";
            }

            if (kit.authlevel > 0)
                if (!session.IsAdmin) return "Nie masz poziomu, aby korzystać z tego zestawu";

            if (!string.IsNullOrEmpty(kit.permission))
                if (!permission.UserHasPermission(session.SteamId.ToString(), kit.permission))
                    return $"Nie masz uprawnien do tego zestawu.";

            var kitData = GetKitData(session.SteamId.m_SteamID, kitname);

            if (kit.max > 0)
                if (kitData.max >= kit.max) return $"Już wykorzystałeś wszystkie zestawy. <i>({kitname})</i>";

            if (kit.cooldown > 0)
            {
                var ct = Convert.ToInt32(CurrentTime());

                if (kitData.cooldown > ct && kitData.cooldown != 0.0)
                {
                    var cd = Convert.ToInt32(kitData.cooldown);

                    if (cd > ct && cd != 0.0)
                    {
                        int a;
                        int t;
                        int s = cd - ct;
                        string sta = "";

                        a = s / 3600;
                        t = s % 3600;
                        if (t != 0)
                            t /= 60;
                        if (s < 3600)
                        {
                            sta = $"Aby dostać zestaw, musisz poczekać {t} minut/y";
                            return $"{sta}";
                        }
                        if (s >= 3600 && t != 0)
                        {
                            sta = $"Aby dostać zestaw, musisz poczekać {a} godzin/y {t} minut/y";
                        }
                        else
                        {
                            sta = $"Aby dostać zestaw, musisz poczekać {a} godzin/y ";
                        }
                        return $"{sta}";
                    }
                }
            }
            return true;
        }

        //////////////////////////////////////////////////////////////////////////////////////
        // Kit Class
        //////////////////////////////////////////////////////////////////////////////////////

        class KitItem
        {
            public int itemid;
            public int amount;
            public int slot;
        }

        class Kit
        {
            public string name;
            public string description;
            public int max;
            public double cooldown;
            public int authlevel;
            public bool hide;
            public string permission;
            public List<KitItem> items = new List<KitItem>();
        }

        //////////////////////////////////////////////////////////////////////////////////////
        // Data Manager
        //////////////////////////////////////////////////////////////////////////////////////

        void SaveKitsData() => Interface.Oxide.DataFileSystem.WriteObject("Kits_Data", kitsData);

        StoredData storedData;
        Dictionary<ulong, Dictionary<string, KitData>> kitsData;

        class StoredData
        {
            public Dictionary<string, Kit> Kits = new Dictionary<string, Kit>();
        }

        class KitData
        {
            public int max;
            public double cooldown;
        }

        void ResetData()
        {
            kitsData.Clear();
            SaveKitsData();
        }

        void Unload() => SaveKitsData();

        void OnServerSave() => SaveKitsData();

        void SaveKits() => Interface.Oxide.DataFileSystem.WriteObject("Kits", storedData);

        void LoadData()
        {
            var kits = Interface.Oxide.DataFileSystem.GetFile("Kits");

            try
            {
                kits.Settings.NullValueHandling = NullValueHandling.Ignore;
                storedData = kits.ReadObject<StoredData>();
            }
            catch
            {
                storedData = new StoredData();
            }
            kits.Settings.NullValueHandling = NullValueHandling.Include;
        }

        KitData GetKitData(ulong userID, string kitname)
        {
            Dictionary<string, KitData> kitDatas;

            if (!kitsData.TryGetValue(userID, out kitDatas)) kitsData[userID] = kitDatas = new Dictionary<string, KitData>();

            KitData kitData;

            if (!kitDatas.TryGetValue(kitname, out kitData)) kitDatas[kitname] = kitData = new KitData();
            return kitData;
        }

        //////////////////////////////////////////////////////////////////////////////////////
        // Kit Editor
        //////////////////////////////////////////////////////////////////////////////////////

        readonly Dictionary<ulong, string> kitEditor = new Dictionary<ulong, string>();

        //////////////////////////////////////////////////////////////////////////////////////
        // Console Command
        //////////////////////////////////////////////////////////////////////////////////////

        List<PlayerSession> FindPlayer(string arg)
        {
            var listPlayers = new List<PlayerSession>();

            ulong steamid;
            ulong.TryParse(arg, out steamid);
            var lowerarg = arg.ToLower();

            foreach (var pair in GameManager.Instance.GetSessions())
            {
                var session = pair.Value;

                if (!session.IsLoaded) continue;
                if (steamid != 0L)
                    if (session.SteamId.m_SteamID == steamid)
                    {
                        listPlayers.Clear();
                        listPlayers.Add(session);
                        return listPlayers;
                    }

                var lowername = session.Name.ToLower();

                if (lowername.Contains(lowerarg)) listPlayers.Add(session);
            }
            return listPlayers;
        }

        //////////////////////////////////////////////////////////////////////////////////////
        // Chat Command
        //////////////////////////////////////////////////////////////////////////////////////

        bool HasAccess(PlayerSession session) => session.IsAdmin || permission.UserHasPermission(session.SteamId.ToString(), permControl);

        void SendListKitEdition(PlayerSession session)
        {
            hurt.SendChatMessage(session, null, "permission \"permission name\" - Uprawnienie");
            hurt.SendChatMessage(session, null, "description \"description text here\" - Opis zestawu");
            hurt.SendChatMessage(session, null, "authlevel XXX - Level uprawnień (Czy mają poisiadać zestaw tylko administratorzy 1 = tak)");
            hurt.SendChatMessage(session, null, "cooldown XXX - Odstęp czasowy");
            hurt.SendChatMessage(session, null, "max XXX - Maksymalna ilość użyć");
            hurt.SendChatMessage(session, null, "items - Dodanie przedmiotów do zestawu");
            hurt.SendChatMessage(session, null, "hide TRUE/FALSE - Ustawić zestaw dla wszystkich graczy?");
        }

        [ChatCommand("kit")]
        void cmdKit(PlayerSession session, string command, string[] args)
        {
            var canTPHook = Interface.Oxide.CallHook("canRedeemKit", session);

            if (canTPHook != null)
            {
                return;
            }

            if (args.Length == 0)
            {
                hurt.SendChatMessage(session, null, "<color=#0066cc>Dostepne zestawy: </color>");

                int s = 0;
                var reason = string.Empty;

                foreach (var pair in storedData.Kits)
                {
                    s += 1;

                    var cansee = CanSeeKit(session, pair.Key, out reason);

                    if (!cansee && string.IsNullOrEmpty(reason)) continue;
                    hurt.SendChatMessage(session, null, $" <b><color=#0066cc>{pair.Value.name}</color></b> - {pair.Value.description} {reason}");
                }
                return;
            }
            if (args.Length == 1)
            {
                int s = 0;

                switch (args[0])
                {
                    case "help":
                        hurt.SendChatMessage(session, null, "Dostepne komendy:");
                        hurt.SendChatMessage(session, null, "/kit - Dostepne zestawy");
                        hurt.SendChatMessage(session, null, "/kit <zestaw> - Otrzymujesz zestaw");
                        if (!HasAccess(session)) return;
                        hurt.SendChatMessage(session, null, "Dostępne komendy administratora");
                        hurt.SendChatMessage(session, null, "/kit add <zestaw> - Dodawanie zestawu");
                        hurt.SendChatMessage(session, null, "/kit remove <zestaw> - Usunięcie zestawu");
                        hurt.SendChatMessage(session, null, "/kit edit <zestaw> - Edytowanie zestawu");
                        hurt.SendChatMessage(session, null, "/kit list - Lista wszystkich zestawów");
                        hurt.SendChatMessage(session, null, "/kit give <nick/steamid> <zestaw> - Dawanie zestawu");
                        hurt.SendChatMessage(session, null, "/kit resetkits - Usunięcie wszystkich zestawow!");
                        hurt.SendChatMessage(session, null, "/kit resetdata - Usunięcie oczekiwania na zestawy i ilość użyć");
                        break;
                    case "add":
                    case "remove":
                    case "edit":
                        if (!HasAccess(session)) { hurt.SendChatMessage(session, null, "Nie masz uprawnien do tej komendy."); return; }
                        hurt.SendChatMessage(session, null, $"/kit {args[0]} <zestaw>");
                        break;
                    case "give":
                        if (!HasAccess(session)) { hurt.SendChatMessage(session, null, "Nie masz uprawnien do tej komendy."); return; }
                        hurt.SendChatMessage(session, null, "/kit give Gracz/SteamID <zestaw>");
                        break;
                    case "list":
                        if (!HasAccess(session)) { hurt.SendChatMessage(session, null, "Nie masz uprawnien do tej komendy."); return; }
                        foreach (var kit in storedData.Kits.Values) { s += 1; hurt.SendChatMessage(session, null, $"<b>{s}</b>{kit.name} - {kit.description}"); }
                        break;
                    case "items":
                        break;
                    case "resetkits":
                        if (!HasAccess(session)) { hurt.SendChatMessage(session, null, "Nie masz uprawnien do tej komendy."); return; }
                        storedData.Kits.Clear();
                        kitEditor.Clear();
                        ResetData();
                        SaveKits();
                        hurt.SendChatMessage(session, null, "Zresetowano wszystkie zestawy!");
                        break;
                    case "resetdata":
                        if (!HasAccess(session)) { hurt.SendChatMessage(session, null, "Nie masz uprawnien do tej komendy."); return; }
                        ResetData();
                        hurt.SendChatMessage(session, null, "Wyczyszczono wszystkie zestawy graczy!");
                        break;
                    default:
                        TryGiveKit(session, args[0].ToLower());
                        break;
                }
                if (args[0] != "items") return;
            }
            if (!HasAccess(session)) { hurt.SendChatMessage(session, null, "Nie masz uprawnien do tej komendy."); return; }

            string kitname;

            switch (args[0])
            {
                case "add":
                    kitname = args[1].ToLower();
                    if (storedData.Kits.ContainsKey(kitname))
                    {
                        hurt.SendChatMessage(session, null, "Ten zestaw juz istnieje.");
                        return;
                    }
                    storedData.Kits[kitname] = new Kit { name = args[1] };
                    kitEditor[session.SteamId.m_SteamID] = kitname;
                    hurt.SendChatMessage(session, null, "Utworzyłeś nowy zestaw: " + args[1]);
                    SendListKitEdition(session);
                    break;
                case "give":
                    if (args.Length < 3)
                    {
                        hurt.SendChatMessage(session, null, "/kit give Gracz/SteamID <zestaw>");
                        return;
                    }
                    kitname = args[2].ToLower();
                    if (!storedData.Kits.ContainsKey(kitname))
                    {
                        hurt.SendChatMessage(session, null, "Ten zestaw nie istnieje");
                        return;
                    }

                    var findPlayers = FindPlayer(args[1]);

                    if (findPlayers.Count == 0)
                    {
                        hurt.SendChatMessage(session, null, "Nie znaleziono graczy.");
                        return;
                    }
                    if (findPlayers.Count > 1)
                    {
                        hurt.SendChatMessage(session, null, "Znaleziono podobnych graczy:");
                        return;
                    }
                    GiveKit(findPlayers[0], kitname);
                    hurt.SendChatMessage(session, null, $"Dałeś graczowi: {findPlayers[0].Name} zestaw: {storedData.Kits[kitname].name}");
                    hurt.SendChatMessage(findPlayers[0], string.Format("Otrzymałeś zestaw {1} od {0}", session.Name, storedData.Kits[kitname].name));
                    break;
                case "edit":
                    kitname = args[1].ToLower();
                    if (!storedData.Kits.ContainsKey(kitname))
                    {
                        hurt.SendChatMessage(session, null, "Ten zestaw nie istnieje");
                        return;
                    }
                    kitEditor[session.SteamId.m_SteamID] = kitname;
                    hurt.SendChatMessage(session, null, $"Edytujesz zestaw: {kitname}");
                    SendListKitEdition(session);
                    break;
                case "remove":
                    kitname = args[1].ToLower();
                    if (!storedData.Kits.Remove(kitname))
                    {
                        hurt.SendChatMessage(session, null, "Ten zestaw nie istnieje");
                        return;
                    }
                    hurt.SendChatMessage(session, null, $"Zestaw <b>{kitname}</b> został usunięty");
                    if (kitEditor[session.SteamId.m_SteamID] == kitname) kitEditor.Remove(session.SteamId.m_SteamID);
                    break;
                default:
                    if (!kitEditor.TryGetValue(session.SteamId.m_SteamID, out kitname))
                    {
                        hurt.SendChatMessage(session, null, "Nie tworzysz ani nie edytujesz zestawu");
                        return;
                    }

                    Kit kit;

                    if (!storedData.Kits.TryGetValue(kitname, out kit))
                    {
                        hurt.SendChatMessage(session, null, "Wystąpił błąd podczas pobierania tego zestawu, czy został zmieniony podczas edycji?");
                        return;
                    }
                    for (var i = 0; i < args.Length; i++)
                    {
                        object editvalue;
                        var key = args[i].ToLower();

                        switch (key)
                        {
                            case "items":
                                kit.items = GetPlayerItems(session);
                                hurt.SendChatMessage(session, null, "Przedmioty z ekwipunku zostały dodane");
                                continue;
                            case "name":
                                continue;
                            case "description":
                                editvalue = kit.description = args[++i];
                                break;
                            case "max":
                                editvalue = kit.max = int.Parse(args[++i]);
                                break;
                            case "cooldown":
                                editvalue = kit.cooldown = double.Parse(args[++i]);
                                break;
                            case "authlevel":
                                editvalue = kit.authlevel = int.Parse(args[++i]);
                                break;
                            case "hide":
                                editvalue = kit.hide = bool.Parse(args[++i]);
                                break;
                            case "permission":
                                editvalue = kit.permission = args[++i];
                                InitializePermissions();
                                break;
                            default:
                                hurt.SendChatMessage(session, null, $"{args[i]} nie jest prawidłowym argumentem");
                                continue;
                        }
                        hurt.SendChatMessage(session, null, $"{key} jest ustanowiony {editvalue ?? "null"}");
                    }
                    break;
            }
            SaveKits();
        }
    }
}
